import PottsPlayground.Test.TestBaseTask
import PottsPlayground.Test.TestColoring
import PottsPlayground.Test.TestBinarized
import PottsPlayground.Test.TestAnneal #uses Coloring task for testing, so should come after TestColoring

#remaining task tests try annealing, so should come after TestAnneal
import PottsPlayground.Test.TestTsp