from PottsPlayground.Tasks.Binarized import Binarized
from PottsPlayground.Tasks.MinorEmbedded import MinorEmbedded
from PottsPlayground.Tasks.TravelingSalesman import TravelingSalesman
from PottsPlayground.Tasks.GraphColoring import GraphColoring
from PottsPlayground.Tasks.Ice40Placer import Ice40Placer
from PottsPlayground.Tasks.BaseTask import BaseTask